# Generated by Django 2.2.2 on 2019-08-08 15:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_education', '0002_observationalert'),
    ]

    operations = [
        migrations.AddField(
            model_name='asyncprocess',
            name='process_type',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



import sys
import logging

import dramatiq
from redis.exceptions import RedisError

from tom_education.models import (
    AsyncError, ASYNC_STATUS_FAILED, PipelineProcess
)

logger = logging.getLogger(__name__)

def task(func, **kwargs):
    """
    Decorator that wraps dramatiq.actor, but runs tasks synchronously during
    tests
    """
    if 'test' not in sys.argv:
        return dramatiq.actor(func, **kwargs)
    func.send = func
    return func


def send_task(task, process, *args):
    """
    Wrapper around queuing a task to start an AsyncProcess sub-class, which
    sets the status and failure message of the process if an exception occurs
    when submitting.

    The task must accept the process's PK as its first argument. *args are
    forwarded to the task.
    """
    try:
        task.send(process.pk, *args)
    except RedisError as ex:
        logger.error('failed to submit job: {}'.format(ex))
        process.status = ASYNC_STATUS_FAILED
        process.failure_message = 'Failed to submit job'
        process.save()


@task
def run_pipeline(process_pk, cls_name):
    """
    Task to run a PipelineProcess sub-class. `cls_name` is the name of the
    pipeline as given in TOM_EDUCATION_PIPELINES.
    """
    try:
        pipeline_cls = PipelineProcess.get_subclass(cls_name)
    except ImportError:
        logger.error('pipeline \'{}\' not found'.format(cls_name), file=sys.stderr)
        return
    try:
        process = pipeline_cls.objects.get(pk=process_pk)
    except pipeline_cls.DoesNotExist:
        logger.error('could not find {} with PK {}'.format(pipeline_cls.__name__, process_pk),
              file=sys.stderr)
        return
    run_process(process)


def run_process(process):
    """
    Helper function to call the run() method of an AsyncProcess, catch errors,
    and update statuses and error messages.

    Note that this runs in the dramatiq worker processes.
    """
    logger.info("running process")
    failure_message = None
    try:
        process.run()
    except AsyncError as ex:
        failure_message = str(ex)
    except NotImplementedError as ex:
        logger.error('S3 not configured correctly: {}'.format(ex))
        failure_message = str(ex)
    except Exception as ex:
        logger.error('unknown error occurred: {}'.format(ex))
        failure_message = 'An unexpected error occurred'

    if failure_message is not None:
        logger.error('task failed: {}'.format(failure_message))
        process.failure_message = failure_message
        process.status = ASYNC_STATUS_FAILED
        process.save()
    logger.error('process finished')



from django import forms
from django.core.exceptions import ValidationError
from tom_dataproducts.models import DataProduct, DataProductGroup

from tom_education.models import ObservationTemplate


def make_templated_form(base_class):
    """
    Return a sub-class of `base_class` which additionally provides links to
    create and instantiate templates.

    `base_class` should be a sub-class of
    tom_observations.facility.GenericObservationForm.
    """
    class F(base_class):
        # Param name for submit button in form, and text for the button
        new_template_action = ('create-template', 'Create new template')

        def __init__(self, *args, **kwargs):
            self.form_url = kwargs.pop('form_url')
            self.show_create = kwargs.pop('show_create')
            super().__init__(*args, **kwargs)

            # If the base form uses crispy, modify the form helper to not
            # include outer <form> tags (since we include these manually in the
            # template)
            if hasattr(self, 'helper'):
                self.helper.form_tag = None

        def get_extra_context(self):
            context = super().get_extra_context()

            # Add template names and instantiation URLs to the context
            templates = ObservationTemplate.objects.filter(
                target__pk=self.initial['target_id'],
                facility=self.initial['facility']
            )
            context['templates'] = []
            for template in templates:
                url = template.get_create_url(self.form_url)
                name = template.name
                context['templates'].append((name, url))

            # Add info for button to create new template
            context['show_new_template_action'] = self.show_create
            context['new_template_action_button'] = self.new_template_action
            return context

    return F


class DataProductSelectionForm(forms.Form):
    """
    Base class for a form including a list of data product checkboxes, where
    the list of data products is taken from the 'products' kwarg to __init__
    """
    def __init__(self, *args, **kwargs):
        products = kwargs.pop('products')
        super().__init__(*args, **kwargs)

        self.product_pks = set([])
        for dp in products:
            str_pk = str(dp.pk)
            self.product_pks.add(str_pk)
            self.fields[str_pk] = forms.fields.BooleanField(required=False)

    def clean(self):
        if not any(self.cleaned_data.get(str_pk) for str_pk in self.product_pks):
            raise ValidationError('No data product selected')

    def get_selected_products(self):
        """
        Return the set of products that were selected
        """
        return {
            DataProduct.objects.get(pk=int(str_pk))
            for str_pk, checked in self.cleaned_data.items() if str_pk in self.product_pks and checked
        }


class DataProductActionForm(DataProductSelectionForm):
    """
    Form for selecting a group of data products from the target page to perform
    some action on them
    """
    action = forms.CharField(required=True)

    def __init__(self, *args, **kwargs):
        target = kwargs.pop('target')
        super().__init__(*args, **kwargs, products=target.dataproduct_set.all())


class GalleryForm(DataProductSelectionForm):
    """
    Form for a user to add a selection of data products to a data product group
    """
    group = forms.ModelChoiceField(DataProductGroup.objects.all())



from dataclasses import dataclass
from datetime import datetime
import json
from typing import Iterable

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied
from django.db.utils import IntegrityError
from django.http import JsonResponse, HttpResponseBadRequest, HttpResponseRedirect, Http404
from django.shortcuts import redirect, reverse
from django.utils.http import urlencode
from django.views.generic import FormView, TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import FormMixin
from tom_dataproducts.models import DataProduct, ObservationRecord, ReducedDatum
from tom_observations.facility import get_service_class
from tom_observations.views import ObservationCreateView
from tom_targets.models import (
    Target, GLOBAL_TARGET_FIELDS, REQUIRED_NON_SIDEREAL_FIELDS,
    REQUIRED_NON_SIDEREAL_FIELDS_PER_SCHEME
)
from tom_targets.views import TargetDetailView, TargetCreateView, TargetUpdateView
from rest_framework.exceptions import NotFound
from rest_framework.generics import CreateAPIView, ListAPIView, RetrieveAPIView
from rest_framework import serializers
from rest_framework.response import Response

from tom_education.forms import make_templated_form, DataProductActionForm, GalleryForm
from tom_education.models import (
    AsyncProcess,
    ASYNC_STATUS_CREATED,
    ObservationAlert,
    ObservationTemplate,
    PipelineProcess,
    TimelapsePipeline,
)
from tom_education.serializers import (
    AsyncProcessSerializer,
    ObservationAlertSerializer,
    PipelineProcessSerializer,
    TargetDetailSerializer,
    TimestampField,
)
from tom_education.tasks import run_pipeline, send_task


class TemplatedObservationCreateView(ObservationCreateView):
    supported_facilities = ('LCO',)

    def get_form_class(self):
        return make_templated_form(super().get_form_class())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Get extra context data from the form object, if applicable
        form = context['form']
        if hasattr(form, 'get_extra_context'):
            context.update(form.get_extra_context())
        context['target'] = self.get_target()
        return context

    def serialize_fields(self, form):
        return json.dumps(form.cleaned_data)

    def form_url(self):
        """
        Return the URL for this form view for the current facility and target
        """
        base = reverse("tom_education:create_obs", kwargs={'facility': self.get_facility()})
        return base + '?' + urlencode({'target_id': self.get_target_id()})

    def can_create_template(self):
        """
        Return True if the current user can create a template for the current
        facility, and False otherwise
        """
        supported_facility = self.get_facility() in self.supported_facilities
        return supported_facility and self.request.user.is_staff

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['form_url'] = self.form_url()
        kwargs['show_create'] = self.can_create_template()
        return kwargs

    def get_initial(self):
        initial = super().get_initial()

        facility = self.get_facility()
        template_id = self.request.GET.get('template_id')
        if template_id:
            template = ObservationTemplate.objects.filter(
                target=self.get_target(),
                facility=facility
            ).get(pk=template_id)

            initial.update(json.loads(template.fields))
            # Set identifier field to something unique based on the template
            id_field = ObservationTemplate.get_identifier_field(facility)
            initial[id_field] = template.get_identifier()

            # Dates need to be converted to just YYYY-MM-DD to display in the
            # widget properly
            for field in ObservationTemplate.get_date_fields(facility):
                dt = initial[field]
                initial[field] = datetime.fromisoformat(dt).strftime('%Y-%m-%d')

        return initial

    def form_valid(self, form):
        facility = self.get_facility()
        if self.get_form_class().new_template_action[0] in form.data:
            if not self.can_create_template():
                raise PermissionDenied()

            # Create new template
            # TODO: deal with None below
            name = form.cleaned_data.get(ObservationTemplate.get_identifier_field(facility))
            try:
                template = ObservationTemplate.objects.create(
                    name=name,
                    target=self.get_target(),
                    facility=facility,
                    fields=self.serialize_fields(form)
                )
            except IntegrityError:
                form.add_error(None, 'Template name "{}" already in use'.format(name))
                return self.form_invalid(form)

            path = template.get_create_url(self.form_url())
            return redirect(path)

        return super().form_valid(form)


class ActionableTargetDetailView(FormMixin, TargetDetailView):
    """
    Extend the target detail view to add a form to select a group of data
    products and perform an action on them.

    A method `handle_<name>(products, form)` is called to handle the form
    submission, where `<name>` is the value of the action field in the form.
    """
    form_class = DataProductActionForm
    template_name = "tom_targets/target_dataview.html"

    def get_success_url(self):
        return reverse('tom_targets:detail', kwargs={'pk': self.get_object().pk})

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['target'] = self.get_object()
        return kwargs

    def get_context_data(self, *args, **kwargs):
        self.object = self.get_object()
        context = super().get_context_data(*args, **kwargs)
        context['dataproducts_form'] = self.get_form()
        context['pipeline_names'] = sorted(PipelineProcess.get_available().keys())
        context['pipeline_flags'] = {}
        for name in context['pipeline_names']:
            pipeline_cls = PipelineProcess.get_subclass(name)
            if pipeline_cls.flags:
                context['pipeline_flags'][name] = pipeline_cls.flags

        return context

    def post(self, _request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        # Form is not rendered in the template, so add form errors as messages
        # Note: this discards information about which field each error relates
        # to
        for err_list in form.errors.values():
            for err_msg in err_list:
                messages.error(self.request, err_msg)
        return self.form_invalid(form)

    def form_valid(self, form):
        products = form.get_selected_products()
        try:
            method = getattr(self, 'handle_{}'.format(form.data['action']))
        except AttributeError:
            return HttpResponseBadRequest('Invalid action \'{}\''.format(form.data['action']))
        return method(products, form)

    def handle_pipeline(self, products, form):
        try:
            name = form.data['pipeline_name']
        except KeyError:
            return HttpResponseBadRequest('No pipeline_name given')
        try:
            pipeline_cls = PipelineProcess.get_subclass(name)
        except KeyError:
            return HttpResponseBadRequest("Invalid pipeline name '{}'".format(name))

        # Get pipeline-specific flags. Initially set all to False; those
        # present in form data will be set to True
        flags = {f: False for f in pipeline_cls.flags} if pipeline_cls.flags else {}
        for key in form.data:
            prefix = 'pipeline_flag_'
            if not key.startswith(prefix):
                continue
            flag = key[len(prefix):]
            if flag not in flags:
                continue
            flags[flag] = True

        target = self.get_object()
        pipe = pipeline_cls.create_timestamped(target, products, flags)
        send_task(run_pipeline, pipe, name)
        return JsonResponse({'ok': True})

    def handle_view_gallery(self, products, form):
        # Redirect to gallery page with product PKs as GET params
        product_pks = [str(p.pk) for p in products]
        base = reverse('tom_education:gallery')
        url = base + '?' + urlencode({'product_pks': ",".join(product_pks)})
        return redirect(url)

    def handle_delete(self, products, form):
        product_pks = [str(p.pk) for p in products]
        base = reverse('tom_education:delete_dataproducts')
        url = base + '?' + urlencode({'product_pks': ",".join(product_pks)})
        return redirect(url)


class GalleryView(FormView):
    """
    Show thumbnails for a number of data products and allow the user to add a
    selection of them to a data product group
    """
    form_class = GalleryForm
    template_name = 'tom_education/gallery.html'

    def get_pks_string(self):
        """
        Return comma separated string of products PKs from GET or POST params
        """
        if self.request.method == 'GET':
            obj = self.request.GET
        else:
            obj = self.request.POST
        return obj.get('product_pks', '')

    def get_products(self, pks_string):
        try:
            return self._products
        except AttributeError:
            pass

        if pks_string:
            pks = pks_string.split(',')
            self._products = {DataProduct.objects.get(pk=int(pk)) for pk in pks}
        else:
            self._products = set([])
        return self._products

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['products'] = self.get_products(self.get_pks_string())
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Comma separated PK string is required to construct form instance, so
        # must be send in the POST request too. Put it in the context so it
        # can be sent as a hidden field
        pks_string = self.get_pks_string()
        context['product_pks'] = pks_string

        products = self.get_products(pks_string)
        if products:
            context['products'] = products
            context['show_form'] = True
        else:
            messages.error(self.request, 'No data products provided')
        return context

    def form_valid(self, form):
        selected = form.get_selected_products()
        group = form.cleaned_data['group']
        for product in selected:
            product.group.add(group)
            product.save()

        # Redirect to group detail view
        msg = 'Added {} data products to group \'{}\''.format(len(selected), group.name)
        messages.success(self.request, msg)
        url = reverse('tom_dataproducts:group-detail', kwargs={'pk': group.pk})
        return HttpResponseRedirect(url)


class AsyncStatusApi(ListAPIView):
    """
    View that finds all AsyncProcess objects associated with a specified Target
    and returns the listing in a JSON response
    """
    serializer_class = AsyncProcessSerializer

    def get_queryset(self):
        try:
            target = Target.objects.get(pk=self.kwargs['target'])
        except Target.DoesNotExist:
            raise Http404
        return AsyncProcess.objects.filter(target=target).order_by('-created')

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer(queryset, many=True)
        timestamp = TimestampField().to_representation(datetime.now())
        return Response({'timestamp': timestamp, 'processes': serializer.data})


class PipelineProcessDetailView(DetailView):
    model = PipelineProcess

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.object.target:
            context['target_url'] = reverse('tom_targets:detail', kwargs={'pk': self.object.target.pk})
        return context


class PipelineProcessApi(RetrieveAPIView):
    """
    Return information about a PipelineProcess in a JSON response
    """
    queryset = PipelineProcess.objects.all()
    serializer_class = PipelineProcessSerializer


@dataclass
class TargetDetailApiInfo:
    """
    Wrapper object containing a target and its timelapses, for serialization
    for the target detail API
    """
    target: Target
    timelapses: Iterable[TimelapsePipeline]


class TargetDetailApiView(RetrieveAPIView):
    """
    Return information about a target and its timelapses, and return a JSON
    response
    """
    serializer_class = TargetDetailSerializer
    # Note: we specify a Target queryset to make use of rest_framework methods
    # to retrieve Target model from the PK kwarg in URL, but it is NOT a Target
    # object that will be serialized
    queryset = Target.objects.all()

    def get_object(self):
        target = super().get_object()
        tl_pipelines = TimelapsePipeline.objects.filter(
            target=target, status=ASYNC_STATUS_CREATED
        ).order_by('-terminal_timestamp')
        return TargetDetailApiInfo(target=target, timelapses=tl_pipelines)


class ObservationAlertApiCreateView(CreateAPIView):
    """
    Create an ObservationAlert by instantiating an ObservationTemplate for a
    given target
    """
    throttle_scope = 'observe'

    serializer_class = ObservationAlertSerializer

    def perform_create(self, serializer):
        data = serializer.validated_data
        try:
            target = Target.objects.get(pk=data['target'])
            facility_class = get_service_class(data['facility'])
            template = ObservationTemplate.objects.get(
                target=target,
                name=data['template_name'],
                facility=data['facility']
            )
        except Target.DoesNotExist:
            raise NotFound(detail='Target not found.')
        except ImportError:
            raise NotFound(detail='Facility not found.')
        except ObservationTemplate.DoesNotExist:
            err = "Template '{}' not found for target '{}' and facility '{}'".format(
                data['template_name'], target.name, data['facility']
            )
            raise NotFound(detail=err)

        # Construct form for creating an observation
        form_data = {
            'target_id': target.pk,
            'facility': facility_class.name
        }
        form_data.update(json.loads(template.fields))
        id_field = ObservationTemplate.get_identifier_field(facility_class.name)
        form_data[id_field] = template.get_identifier()
        form_data.update(data.get('overrides', {}))
        form = facility_class.get_form(None)(form_data)  # observation type is not relevant to us
        if not form.is_valid():
            raise serializers.ValidationError(form.errors)

        # Submit observation using facility class
        observation_ids = facility_class().submit_observation(form.observation_payload())
        assert len(observation_ids) == 1, (
            'Submittion created multiple observation IDs: {}'.format(observation_ids)
        )
        # Create Observation record and alert
        ob = ObservationRecord.objects.create(
            target=target,
            facility=facility_class.name,
            parameters=form.serialize_parameters(),
            observation_id=observation_ids[0]
        )
        ObservationAlert.objects.create(email=data['email'], observation=ob)


class DataProductDeleteMultipleView(LoginRequiredMixin, TemplateView):
    template_name = 'tom_education/dataproduct_confirm_delete_multiple.html'

    def get_products(self, pks_string):
        pks = pks_string.split(',')
        return {DataProduct.objects.get(pk=int(pk)) for pk in pks}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.request.method == 'GET':
            context['next'] = self.request.META.get('HTTP_REFERER', '/')
            context['product_pks'] = self.request.GET.get('product_pks', '')
            context['to_delete'] = self.get_products(context['product_pks'])
        return context

    def post(self, request, *args, **kwargs):
        prods = self.get_products(self.request.POST.get('product_pks', []))
        for prod in prods:
            ReducedDatum.objects.filter(data_product=prod).delete()
            prod.data.delete()
            prod.delete()
        messages.success(request, 'Deleted {} data products'.format(len(prods)))
        return HttpResponseRedirect(self.request.POST.get('next', '/'))


class NonSiderealFieldsMixin:
    """
    Mixin for views which adds information to the template context about the
    required fields per scheme for non-sidereal targets. This allows client
    side JS to hide fields which are not applicable for the selected scheme.

    Relies on the view having a method get_target_type() which returns
    Target.SIDEREAL or Target.NON_SIDEREAL
    """
    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        if self.get_target_type() == Target.NON_SIDEREAL:
            form = self.get_form()

            # Build list of base fields that should always be shown, including
            # non-model fields declared in the form itself and extra fields
            declared = list(form.declared_fields.keys())
            extra = list(getattr(form, 'extra_fields', {}).keys())
            base = GLOBAL_TARGET_FIELDS + REQUIRED_NON_SIDEREAL_FIELDS + declared + extra

            context['non_sidereal_fields'] = json.dumps({
                'base_fields': base,
                'scheme_fields': REQUIRED_NON_SIDEREAL_FIELDS_PER_SCHEME,
            })
        return context


class EducationTargetCreateView(NonSiderealFieldsMixin, TargetCreateView):
    pass


class EducationTargetUpdateView(NonSiderealFieldsMixin, TargetUpdateView):
    def get_target_type(self):
        return self.object.type



from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand
from tom_dataproducts.models import DataProduct
from tom_observations.facility import get_service_class

from tom_education.constants import RAW_FILE_EXTENSION
from tom_education.models import ObservationAlert, TimelapsePipeline, TIMELAPSE_TAG


class Command(BaseCommand):
    help = ('Update the status of observations with alerts, download any new '
            'data, create timelapses, and send notification emails.')

    IMAGE_FILE_SUFFIX = '.fits.fz'

    def handle(self, *args, **options):
        # Check required settings are present before proceeding
        try:
            from_addr = settings.TOM_EDUCATION_FROM_EMAIL_ADDRESS
        except AttributeError:
            self.stderr.write('TOM_EDUCATION_FROM_EMAIL_ADDRESS not set in settings.py')
            return

        # Keep track of targets and alerts with new data
        new_data_targets = set([])
        new_data_alerts = set([])

        for alert in ObservationAlert.objects.all():
            ob = alert.observation
            facility_class = get_service_class(ob.facility)
            facility = facility_class()

            if ob.status not in facility.get_terminal_observing_states():
                facility.update_observation_status(ob.observation_id)
                self.stdout.write('Checking for new data for observation {}'.format(ob.observation_id))
                if facility.save_data_products(ob):
                    new_data_targets.add(ob.target)
                    new_data_alerts.add(alert)

        for target in new_data_targets:
            # Make new timelapse
            # TODO: supress imageio output
            prods = (target.dataproduct_set.filter(data__endswith=self.IMAGE_FILE_SUFFIX)
                                           .exclude(data__endswith=RAW_FILE_EXTENSION))
            if not prods.exists():
                continue
            new_pipeline = TimelapsePipeline.create_timestamped(target, prods)
            self.stdout.write('Creating timelapse for target {}'.format(target.name))
            try:
                new_pipeline.run()
            except AsyncError as ex:
                self.stderr.write(f'Failed to create timelapse: {ex}')
                return
            new_tl = new_pipeline.group.dataproduct_set.first()

            # Delete old timelapses
            # TODO: control deletion from settings.py
            timelapses = (DataProduct.objects
                                     .filter(target=target, tag=TIMELAPSE_TAG)
                                     .exclude(pk=new_tl.pk)
                                     .all())
            for tl in timelapses:
                tl.delete()
                tl.data.delete(save=False)

        for alert in new_data_alerts:
            target = alert.observation.target
            # Send update email
            subject = "Observation for '{}' has new data".format(target.name)
            message = ("Your observation for '{}' has completed, and a "
                       "timelapse is available".format(target.name))
            to = [alert.email]
            send_mail(subject, message, from_addr, to)
        self.stdout.write('Send {} email updates'.format(len(new_data_alerts)))



import os.path

from django.shortcuts import reverse
from rest_framework import serializers
from tom_targets.models import Target

from tom_education.models import AsyncProcess, PipelineProcess


class TimestampField(serializers.Field):
    def to_representation(self, dt):
        return dt.timestamp()


class AsyncProcessSerializer(serializers.ModelSerializer):
    created = TimestampField()
    terminal_timestamp = TimestampField()
    failure_message = serializers.SerializerMethodField()
    view_url = serializers.SerializerMethodField()

    class Meta:
        model = AsyncProcess
        fields = [
            'identifier', 'created', 'status', 'terminal_timestamp', 'failure_message', 'view_url',
            'process_type'
        ]

    def get_view_url(self, obj):
        """
        Special case for PipelineProcess objects: provide link to detail view
        """
        if hasattr(obj, 'pipelineprocess'):
            return reverse('tom_education:pipeline_detail', kwargs={'pk': obj.pk})
        return None

    def get_failure_message(self, obj):
        return obj.failure_message or None


class PipelineProcessSerializer(AsyncProcessSerializer):
    group_name = serializers.SerializerMethodField()
    group_url = serializers.SerializerMethodField()
    logs = serializers.SerializerMethodField()

    class Meta:
        model = PipelineProcess
        fields = [
            'identifier', 'created', 'status', 'terminal_timestamp', 'failure_message', 'view_url',
            'logs', 'group_name', 'group_url'
        ]

    def get_group_name(self, obj):
        if obj.group:
            return obj.group.name
        return None

    def get_group_url(self, obj):
        if obj.group:
            return reverse('tom_dataproducts:group-detail', kwargs={'pk': obj.group.pk})
        return None

    def get_logs(self, obj):
        """
        Make sure logs is always a string
        """
        return obj.logs or ''


class TargetSerializer(serializers.ModelSerializer):
    """
    Serialize a subset of the Target fields, plus any extra fields
    """
    class Meta:
        model = Target
        fields = ['name', 'extra_fields']


class TimelapsePipelineSerializer(serializers.Serializer):
    """
    Serialize basic info for a timelapse from a TimelapsePipeline object,
    including the (relative) URL to the actual timelapse file
    """
    name = serializers.SerializerMethodField()
    format = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()
    created = serializers.SerializerMethodField()
    frames = serializers.SerializerMethodField()

    def _get_dataproduct(self, obj):
        return obj.group.dataproduct_set.first()

    def get_name(self, obj):
        return os.path.basename(self._get_dataproduct(obj).data.name)

    def get_format(self, obj):
        filename = self.get_name(obj)
        return filename.split('.')[-1]

    def get_url(self, obj):
        return self._get_dataproduct(obj).data.url

    def get_created(self, obj):
        return TimestampField().to_representation(obj.terminal_timestamp)

    def get_frames(self, obj):
        return obj.input_files.count()


class TargetDetailSerializer(serializers.Serializer):
    """
    Response for target detail API: includes information about the target and
    its timelapses
    """
    target = TargetSerializer()
    timelapses = serializers.ListSerializer(child=TimelapsePipelineSerializer())


class ObservationAlertSerializer(serializers.Serializer):
    target = serializers.IntegerField(min_value=1)
    template_name = serializers.CharField()
    facility = serializers.CharField()
    overrides = serializers.DictField(required=False)
    email = serializers.EmailField()



from datetime import datetime
from urllib.parse import urlparse, parse_qs, urlunparse

from django.db import models
from django.utils.http import urlencode
from tom_targets.models import Target


class ObservationTemplate(models.Model):
    name = models.CharField(max_length=255, null=False)
    target = models.ForeignKey(Target, on_delete=models.CASCADE, null=False)
    facility = models.CharField(max_length=255, null=False)
    # Form fields serialized as a JSON string
    fields = models.TextField()

    class Meta:
        unique_together = ('name', 'target', 'facility')

    def get_create_url(self, base_url):
        """
        Return URL for instantiating this template by adding 'template_id' GET
        parameter to base create URL
        """
        # Need to parse base URL and combine GET parameters
        parsed_url = urlparse(base_url)
        params = parse_qs(parsed_url.query)
        for key, val in params.items():
            params[key] = val[0]
        params['template_id'] = self.pk
        parts = list(parsed_url)
        parts[4] = urlencode(params)
        return urlunparse(parts)

    def get_identifier(self):
        """
        Return an identifier for an instantiation of this template, based on
        the template name and current date and time
        """
        now = datetime.now()
        fmt = '%Y-%m-%d-%H%M%S'
        return "{}-{}".format(self.name, now.strftime(fmt))

    @staticmethod
    def get_identifier_field(facility):
        """
        Return name of the field used to extract template name when creating a
        template. This field is also used to store the identifier for
        instantiated templates
        """
        if facility == 'LCO':
            return 'name'
        raise NotImplementedError

    @staticmethod
    def get_date_fields(facility):
        """
        Return a sequence of field names whose type is datetime
        """
        if facility == 'LCO':
            return ['start', 'end']
        return []



import json

from django import template
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from tom_dataproducts.models import DataProductGroup
from tom_targets.models import TargetExtra
from tom_observations.models import ObservationRecord

from tom_education.constants import RAW_FILE_EXTENSION

register = template.Library()


@register.inclusion_tag('tom_education/dataproduct_checkbox.html')
def dataproduct_checkbox(product):
    is_reduced = 'fits' in product.data.name and not product.data.name.endswith(RAW_FILE_EXTENSION)
    return {
        'product': product,
        'reduced': is_reduced,
        'groups': product.group.all(),
    }


@register.inclusion_tag('tom_education/dataproduct_selection_buttons.html', takes_context=True)
def dataproduct_selection_buttons(context, show_group_selection=True):
    context['show_group_selection'] = show_group_selection
    if show_group_selection:
        target = context['target']
        context['data_product_groups'] = DataProductGroup.objects.filter(dataproduct__target=target).distinct()
    return context


@register.simple_tag
def loading_message():
    return 'Loading...'


@register.simple_tag
def status_icon(status):
    EMOJI_CHOICES = {
        'PENDING': '🕒',
        'FAILED': '❌',
        'WINDOW_EXPIRED': '❌',
        'COMPLETED': '✅'
    }
    return format_html(mark_safe('<span title={}>{}</span>'.format(
        status, EMOJI_CHOICES.get(status, '😑')
    )))


@register.inclusion_tag('tom_targets/partials/target_features.html')
def featured_images(target):
    return {'target': target}


@register.inclusion_tag('tom_dataproducts/partials/dataproduct_thumbs_for_target.html')
def dataproduct_thumbs_for_target(target):
    return {
        'products': target.dataproduct_set.filter(data__contains='fits').order_by('-created')[:6],
        'target': target
    }


@register.inclusion_tag('tom_dataproducts/partials/dataproduct_other_for_target.html')
def dataproduct_other_for_target(target):
    timelapses = target.dataproduct_set.filter(tag='timelapse')
    return {
        'timelapse': timelapses.latest() if timelapses else None,
        'photometry': target.dataproduct_set.filter(tag='photometry'),
        'target': target
    }


@register.filter
def dataproduct_extrainfo(text):
    try:
        return json.loads(text)
    except json.JSONDecodeError:
        return None


@register.inclusion_tag('tom_targets/partials/target_latest_active.html')
def target_latest_active():
    return {'targets': TargetExtra.objects.filter(key='active', value=True)}


@register.inclusion_tag('tom_observations/partials/observationrecord_latest.html')
def observationrecord_latest(value=10):
    value = int(value)
    return {'observationrecords': ObservationRecord.objects.all().order_by('-modified')[:value]}


@register.filter
def get_form_field(form, label):
    """
    Return a bound field with the given label from a form
    """
    return form.fields[label].get_bound_field(form, label)



from django.db import models

from tom_observations.models import ObservationRecord


class ObservationAlert(models.Model):
    """
    A model to store an email address with an observation, so that an alert can
    be issued when there is a change in the observation status
    """
    observation = models.ForeignKey(ObservationRecord, on_delete=models.CASCADE)
    email = models.EmailField()



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='tom_education',
    version='1.1.0',
    description='TOM toolkit plugin for educational projects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Joe Singleton',
    author_email='joesingo@gmail.com',
    maintainer='Edward Gomez',
    maintainer_email='egomez@lco.global',
    install_requires=[
        'astroscrappy==1.0.8',
        'tomtoolkit==1.0.0',
        'numpy',
        'imageio-ffmpeg==0.3.0',
        'imageio==2.5.0',
        'django-dramatiq==0.7.1',
        'dramatiq==1.6.0',
        'redis==3.2.1',
        'watchdog==0.9.0',
        'watchdog-gevent==0.1.1',
        'djangorestframework==3.10.1',
    ],
    packages=find_packages(),
    include_package_data=True,
    extras_require={
        'test': ['factory_boy', 'rise-set']
    }
)



# Generated by Django 2.2.2 on 2019-07-30 15:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0003_auto_20190503_2318'),
        ('tom_education', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ObservationAlert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254)),
                ('observation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_observations.ObservationRecord')),
            ],
        ),
    ]



import json
import operator

import requests
from django import forms
from django.conf import settings
from django.core.files.base import ContentFile
from dateutil.parser import parse
from tom_observations.facility import get_service_class
from tom_observations.facilities.lco import LCOFacility, LCOImagingObservationForm, make_request, PORTAL_URL

try:
    AUTO_THUMBNAILS = settings.AUTO_THUMBNAILS
except AttributeError:
    AUTO_THUMBNAILS = False


class EducationLCOForm(LCOImagingObservationForm):
    EXPOSURE_FIELDS = {
        'exposure_count': (forms.IntegerField, {'min_value': 1}),
        'exposure_time': (forms.FloatField, {'min_value': 0.1}),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['filter']
        del self.fields['exposure_count']
        del self.fields['exposure_time']

        # Keep a list of (filter_code, filter name, [field_key, ...]) for all
        # available filters and the exposure fields for them
        self.filter_fields = []

        # Add each of the exposure fields for each filter (ordered
        # alphabetically by filter name)
        for filter_code, filter_name in sorted(self.filter_choices(), key=operator.itemgetter(1)):
            keys = []
            for exp_field_name, (exp_field_class, kwargs) in self.EXPOSURE_FIELDS.items():
                key = f'{filter_code}_{exp_field_name}'
                self.fields[key] = exp_field_class(required=False, **kwargs)
                keys.append(key)
            self.filter_fields.append((filter_code, filter_name, keys))

    def is_valid(self):
        if not self.is_bound or self.errors:  # Note: accessing errors calls clean()
            return False

        # Validation below is copied from tom_base. Note that calling
        # super().is_valid() is not desirable here since the tom_base code
        # *always* validates the observation (which involves an API call) even
        # if clean() has already thrown up errors
        obs_module = get_service_class(self.cleaned_data['facility'])
        errors = obs_module().validate_observation(self.observation_payload())
        if errors:
            self.add_error(None, self._flatten_error_dict(errors))
        return not errors

    def clean(self):
        """
        Validate the exposure count/time fields
        """
        num_filters = 0
        error_messages = []

        for code, name, _ in self.filter_fields:
            time_key = f'{code}_exposure_time'
            count_key = f'{code}_exposure_count'
            time = self.cleaned_data.get(time_key)
            count = self.cleaned_data.get(count_key)
            # Remove fields from data dict if neither have been specified
            if not (time or count):
                del self.cleaned_data[time_key]
                del self.cleaned_data[count_key]
                continue
            # It is an error if only one is specified
            if not count:
                error_messages.append(f"Exposure count missing for filter '{name}'")
                continue
            elif not time:
                error_messages.append(f"Exposure time missing for filter '{name}'")
                continue
            else:
                num_filters += 1

        if error_messages:
            raise forms.ValidationError(error_messages)
        if num_filters == 0:
            raise forms.ValidationError('No filters selected')
        return super().clean()

    @staticmethod
    def get_schedulable_codes(api_response):
        """
        For a JSON response from the instruments API, return a dictionary
        mapping instrument names to lists of filter/slit codes which are
        schedulable
        """
        info = {}
        for name, val in api_response.items():
            keys = ('filters', 'slits')
            for key in keys:
                objs = val['optical_elements'].get(key, [])
                allowed = [obj['code'] for obj in objs if obj['schedulable']]
                if name not in info:
                    info[name] = []
                info[name] += allowed
        return info

    def get_extra_context(self):
        """
        Provide extra context to the view using this form.
        """
        json_response = make_request('GET', PORTAL_URL + '/api/instruments/').json()
        info = EducationLCOForm.get_schedulable_codes(json_response)
        return {
            'instrument_filters': json.dumps(info),
            'filter_fields': self.filter_fields
        }

    def observation_payload(self):
        """
        Modify observation payload for LCO API to add multiple 'configuration'
        objects to the 'request' object, so that we can have multiple filters
        and exposure settings.

        Note: ideally we would provide only a single 'configuration' which
        contains multiple 'instrument_config' objects -- this would prevent
        duplication of instrument/target/constraint fields in each
        configuration -- but the LCO API does not currently support this.
        """
        payload = super().observation_payload()
        payload['requests'][0]['configurations'] = list(self._build_configurations())
        return payload

    def _build_configuration(self):
        # observation_payload() in the parent class calls this method to build
        # configurations, but the implementation in the parent class only works
        # for a single filter, and causes an exception with this form since the
        # 'filter' field has been removed. Override it here to avoid an error
        # in observation_payload().
        return None

    def _build_configurations(self):
        """
        Generator which yields dicts for the 'configurations' key of the
        'request' object in the observation payload
        """
        for filter_code, _, _ in self.filter_fields:
            try:
                instrument_config = {
                    'exposure_count': self.cleaned_data[f'{filter_code}_exposure_count'],
                    'exposure_time': self.cleaned_data[f'{filter_code}_exposure_time'],
                    'optical_elements': {'filter': filter_code}
                }
            except KeyError:
                continue

            yield {
                'type': self.instrument_to_type(self.cleaned_data['instrument_type']),
                'instrument_type': self.cleaned_data['instrument_type'],
                'target': self._build_target_fields(),
                'instrument_configs': [instrument_config],
                'acquisition_config': {},
                'guiding_config': {},
                'constraints': {
                    'max_airmass': self.cleaned_data['max_airmass']
                }
            }


class EducationLCOFacility(LCOFacility):
    def get_form(self, *args):
        return EducationLCOForm

    def data_products(self, observation_id, product_id=None):
        """
        Override this method to include reduction level in the dict for each
        data product
        """
        products = []
        for frame in self._archive_frames(observation_id, product_id):
            extra = {
                'date_obs': frame['DATE_OBS'],
                'instrument': frame['INSTRUME'],
                'siteid': frame['SITEID'],
                'telid': frame['TELID'],
                'exp_time': frame['EXPTIME'],
                'filter': frame['FILTER']
            }
            products.append({
                'id': frame['id'],
                'filename': frame['filename'],
                'created': parse(frame['DATE_OBS']),
                'url': frame['url'],
                'reduced': frame['RLEVEL'] == 91,
                'extra': extra
            })
        return products

    def save_data_products(self, observation_record, product_id=None):
        from tom_dataproducts.models import DataProduct
        from tom_dataproducts.utils import create_image_dataproduct
        final_products = []
        products = self.data_products(observation_record.observation_id, product_id)
        for product in products:
            dp, created = DataProduct.objects.get_or_create(
                product_id=product['id'],
                target=observation_record.target,
                observation_record=observation_record,
                extra_data=json.dumps(product['extra'])
            )
            if created:
                product_data = requests.get(product['url']).content
                dfile = ContentFile(product_data)
                dp.data.save(product['filename'], dfile)
                dp.save()
                dp.get_preview()
            if AUTO_THUMBNAILS:
                create_image_dataproduct(dp)
            final_products.append(dp)
        return final_products



import logging

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from tom_dataproducts.models import DataProductGroup
from tom_targets.models import Target

from tom_education.constants import RAW_FILE_EXTENSION
from tom_education.models import TimelapsePipeline, AsyncError


# Supress imageio warnings
logging.getLogger("imageio").setLevel(logging.CRITICAL)


class Command(BaseCommand):
    help = 'Create a timelapse from a series of data products for a target'

    def add_arguments(self, parser):
        parser.add_argument('target_pk', type=int)

    def handle(self, *args, **options):
        # Get the target
        target_pk = options['target_pk']
        try:
            target = Target.objects.get(pk=target_pk)
        except Target.DoesNotExist:
            raise CommandError('Target \'{}\' does not exist'.format(target_pk))

        # Get the 'good quality' group
        group_name = settings.TOM_EDUCATION_TIMELAPSE_GROUP_NAME
        try:
            group, _ = DataProductGroup.objects.get_or_create(name=group_name)
        except DataProductGroup.MultipleObjectsReturned:
            raise CommandError(
                'Multiple data products groups found with name \'{}\''.format(group_name)
            )

        prods = group.dataproduct_set.filter(target=target).exclude(data__endswith=RAW_FILE_EXTENSION)
        self.stdout.write("Creating timelapse of {n} files for target '{name}'...".format(
            n=prods.count(),
            name=target.name
        ))

        if not prods.exists():
            self.stdout.write('Nothing to do')
            return

        pipe = TimelapsePipeline.create_timestamped(target, prods)
        try:
            pipe.run()
        except AsyncError as ex:
            self.stderr.write(f'Failed to create timelapse: {ex}')
        else:
            prod = pipe.group.dataproduct_set.first()
            msg = f'Created timelapse {prod.data.path}'
            self.stdout.write(self.style.SUCCESS(msg))
